@@ -2,7 +2,7 @@
  * libdpkg - Debian packaging suite library routines
  * i18n.h - i18n support
  *
- * Copyright © 2008 Guillem Jover <guillem@debian.org>
+ * Copyright © 2008-2011 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,8 +27,14 @@ DPKG_BEGIN_DECLS
 
 #include <gettext.h>
 
+/* We need to include this because pgettext() uses LC_MESSAGES, but libintl.h
+ * which gets pulled by gettext.h only includes it if building optimized. */
+#include <locale.h>
+
 #define _(str) gettext(str)
+#define P_(str, str_plural, n) ngettext(str, str_plural, n)
 #define N_(str) gettext_noop(str)
+#define C_(ctxt, str) pgettext(ctxt, str)
 
 DPKG_END_DECLS
 
