@@ -4,6 +4,7 @@
  *
  * Copyright © 2007 Canonical Ltd
  * Written by Ian Jackson <ian@chiark.greenend.org.uk>
+ * Copyright © 2008-2011 Guillem Jover <guillem@debian.org>
  *
  * This is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,37 +28,40 @@
 
 #include <assert.h>
 #include <errno.h>
+#include <stdlib.h>
 #include <unistd.h>
 
 #include <dpkg/i18n.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
-#include <dpkg/pkg-list.h>
 #include <dpkg/dlist.h>
+#include <dpkg/dir.h>
+#include <dpkg/pkg-spec.h>
+#include <dpkg/trigdeferred.h>
+#include <dpkg/triglib.h>
 
-const char *
-illegal_triggername(const char *p)
+/*========== Recording triggers. ==========*/
+
+static char *triggersdir, *triggersfilefile, *triggersnewfilefile;
+
+static char *
+trig_get_filename(const char *dir, const char *filename)
 {
-	int c;
+	char *path;
 
-	if (!*p)
-		return _("empty trigger names are not permitted");
+	m_asprintf(&path, "%s/%s", dir, filename);
 
-	while ((c = *p++)) {
-		if (c <= ' ' || c >= 0177)
-			return _("trigger name contains invalid character");
-	}
-
-	return NULL;
+	return path;
 }
 
-/*========== recording triggers ==========*/
+static struct trig_hooks trigh;
 
-/*---------- noting trigger activation in memory ----------*/
+/*---------- Noting trigger activation in memory. ----------*/
 
-/* Called via trig_*activate* et al from:
+/*
+ * Called via trig_*activate* et al from:
  *   - trig_incorporate: reading of Unincorp (explicit trigger activations)
- *   - various places: processing start (`activate' in triggers ci file)
+ *   - various places: processing start (‘activate’ in triggers ci file)
  *   - namenodetouse: file triggers during unpack / remove
  *   - deferred_configure: file triggers during config file processing
  *
@@ -66,19 +70,21 @@ illegal_triggername(const char *p)
  * recorded (how would we know?) and (b) we don't enqueue them for
  * deferred processing in this run.
  *
- * We add the trigger to Triggers-Pending first.  This makes it
+ * We add the trigger to Triggers-Pending first. This makes it
  * harder to get into the state where Triggers-Awaited for aw lists
- * pend but Triggers-Pending for pend is empty.  (See also the
+ * pend but Triggers-Pending for pend is empty. (See also the
  * comment in deppossi_ok_found regarding this situation.)
  */
 
-/* aw might be NULL, and trig is not copied! */
+/*
+ * aw might be NULL.
+ * trig is not copied!
+ */
 static void
 trig_record_activation(struct pkginfo *pend, struct pkginfo *aw, const char *trig)
 {
 	if (pend->status < stat_triggersawaited)
-		/* Not interested then. */
-		return;
+		return; /* Not interested then. */
 
 	if (trig_note_pend(pend, trig))
 		modstatdb_note_ifwrite(pend);
@@ -94,60 +100,6 @@ trig_record_activation(struct pkginfo *pend, struct pkginfo *aw, const char *tri
 		}
 }
 
-/* NB that this is also called from fields.c where *pend is a temporary! */
-int
-trig_note_pend_core(struct pkginfo *pend, const char *trig /*not copied!*/)
-{
-	struct trigpend *tp;
-
-	for (tp = pend->trigpend_head; tp; tp = tp->next)
-		if (!strcmp(tp->name, trig))
-			return 0;
-
-	tp = nfmalloc(sizeof(*tp));
-	tp->name = trig;
-	tp->next = pend->trigpend_head;
-	pend->trigpend_head = tp;
-
-	return 1;
-}
-
-/* Returns: 1 for done, 0 for already noted. */
-int
-trig_note_pend(struct pkginfo *pend, const char *trig /*not copied!*/)
-{
-	if (!trig_note_pend_core(pend, trig))
-		return 0;
-
-	pend->status = pend->trigaw.head ? stat_triggersawaited :
-	               stat_triggerspending;
-
-	return 1;
-}
-
-/* Returns: 1 for done, 0 for already noted. */
-/* NB that this is called also from fields.c where *aw is a temporary
- * but pend is from findpackage()! */
-int
-trig_note_aw(struct pkginfo *pend, struct pkginfo *aw)
-{
-	struct trigaw *ta;
-
-	/* We search through aw's list because that's probably shorter. */
-	for (ta = aw->trigaw.head; ta; ta = ta->sameaw.next)
-		if (ta->pend == pend)
-			return 0;
-
-	ta = nfmalloc(sizeof(*ta));
-	ta->aw = aw;
-	ta->pend = pend;
-	ta->nextsamepend = pend->othertrigaw_head;
-	pend->othertrigaw_head = ta;
-	LIST_LINK_TAIL_PART(aw->trigaw, ta, sameaw.);
-
-	return 1;
-}
-
 void
 trig_clear_awaiters(struct pkginfo *notpend)
 {
@@ -158,7 +110,7 @@ trig_clear_awaiters(struct pkginfo *notpend)
 
 	ta = notpend->othertrigaw_head;
 	notpend->othertrigaw_head = NULL;
-	for (; ta; ta = ta->nextsamepend) {
+	for (; ta; ta = ta->samepend_next) {
 		aw = ta->aw;
 		if (!aw)
 			continue;
@@ -171,48 +123,26 @@ trig_clear_awaiters(struct pkginfo *notpend)
 	}
 }
 
-static struct pkg_list *trig_awaited_pend_head;
-
-void
-trig_enqueue_awaited_pend(struct pkginfo *pend)
-{
-	struct pkg_list *tp;
-
-	for (tp = trig_awaited_pend_head; tp; tp = tp->next)
-		if (tp->pkg == pend)
-			return;
-
-	pkg_list_prepend(&trig_awaited_pend_head, pend);
-}
-
 /*
  * Fix up packages in state triggers-awaited w/o the corresponding package
  * with pending triggers. This can happen when dpkg was interrupted
  * while in modstatdb_note, and the package in triggers-pending had its
- * state modified but dpkg could not clearing the awaiters.
+ * state modified but dpkg could not finish clearing the awaiters.
  *
- * XXX: possibly get rid of some of the checks done somewhere else for
+ * XXX: Possibly get rid of some of the checks done somewhere else for
  *      this condition at run-time.
  */
 void
 trig_fixup_awaiters(enum modstatdb_rw cstatus)
 {
-	struct pkg_list *tp;
-
 	if (cstatus < msdbrw_write)
 		return;
 
-	for (tp = trig_awaited_pend_head; tp; tp = tp->next)
-		if (!tp->pkg->trigpend_head)
-			trig_clear_awaiters(tp->pkg);
-
-	pkg_list_free(trig_awaited_pend_head);
-	trig_awaited_pend_head = NULL;
+	trig_awaited_pend_foreach(trig_clear_awaiters);
+	trig_awaited_pend_free();
 }
 
-/*---------- generalised handling of trigger kinds ----------*/
-
-static const struct trigkindinfo tki_explicit, tki_file, tki_unknown;
+/*---------- Generalized handling of trigger kinds. ----------*/
 
 struct trigkindinfo {
 	/* Only for trig_activate_start. */
@@ -221,9 +151,11 @@ struct trigkindinfo {
 	/* Rest are for everyone: */
 	void (*activate_awaiter)(struct pkginfo *pkg /* may be NULL */);
 	void (*activate_done)(void);
-	void (*interest_change)(const char *name, struct pkginfo *pkg, int signum);
+	void (*interest_change)(const char *name, struct pkginfo *pkg, int signum,
+	                        enum trig_options opts);
 };
 
+static const struct trigkindinfo tki_explicit, tki_file, tki_unknown;
 static const struct trigkindinfo *dtki;
 
 /* As passed into activate_start. */
@@ -245,14 +177,15 @@ trig_classify_byname(const char *name)
 		return &tki_file;
 	}
 
-	if (!illegal_packagename(name, NULL) && !strchr(name, '_'))
+	if (!pkg_name_is_illegal(name, NULL) && !strchr(name, '_'))
 		return &tki_explicit;
 
 invalid:
 	return &tki_unknown;
 }
 
-/* Calling sequence is:
+/*
+ * Calling sequence is:
  *  trig_activate_start(triggername)
  *  dtki->activate_awaiter(awaiting_package) } zero or more times
  *  dtki->activate_awaiter(0)                }  in any order
@@ -266,7 +199,7 @@ trig_activate_start(const char *name)
 	dtki->activate_start();
 }
 
-/*---------- unknown trigger kinds ----------*/
+/*---------- Unknown trigger kinds. ----------*/
 
 static void
 trk_unknown_activate_start(void)
@@ -283,12 +216,13 @@ trk_unknown_activate_done(void)
 {
 }
 
-static void
-trk_unknown_interest_change(const char *trig, struct pkginfo *pkg, int signum)
+static void DPKG_ATTR_NORET
+trk_unknown_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+                            enum trig_options opts)
 {
 	ohshit(_("invalid or unknown syntax in trigger name `%.250s'"
 	         " (in trigger interests for package `%.250s')"),
-	       trig, pkg->name);
+	       trig, pkg_describe(pkg, pdo_foreign));
 }
 
 static const struct trigkindinfo tki_unknown = {
@@ -298,7 +232,7 @@ static const struct trigkindinfo tki_unknown = {
 	.interest_change = trk_unknown_interest_change,
 };
 
-/*---------- explicit triggers ----------*/
+/*---------- Explicit triggers. ----------*/
 
 static FILE *trk_explicit_f;
 static struct varbuf trk_explicit_fn;
@@ -318,10 +252,11 @@ trk_explicit_start(const char *trig)
 {
 	trk_explicit_activate_done();
 
-	varbufreset(&trk_explicit_fn);
-	varbufaddstr(&trk_explicit_fn, triggersdir);
-	varbufaddstr(&trk_explicit_fn, trig);
-	varbufaddc(&trk_explicit_fn, 0);
+	varbuf_reset(&trk_explicit_fn);
+	varbuf_add_str(&trk_explicit_fn, triggersdir);
+	varbuf_add_char(&trk_explicit_fn, '/');
+	varbuf_add_str(&trk_explicit_fn, trig);
+	varbuf_end_str(&trk_explicit_fn);
 
 	trk_explicit_f = fopen(trk_explicit_fn.buf, "r");
 	if (!trk_explicit_f) {
@@ -350,6 +285,7 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 	char buf[1024];
 	const char *emsg;
 	struct pkginfo *pend;
+	struct pkg_spec pkgspec = PKG_SPEC_INIT(psf_def_native | psf_no_check);
 
 	if (!trk_explicit_f)
 		return;
@@ -359,58 +295,108 @@ trk_explicit_activate_awaiter(struct pkginfo *aw)
 		        trk_explicit_fn.buf);
 
 	while (trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
-		if ((emsg = illegal_packagename(buf, NULL)))
+		char *slash;
+		bool noawait = false;
+		slash = strchr(buf, '/');
+		if (slash && strcmp("/noawait", slash) == 0) {
+			noawait = true;
+			*slash = '\0';
+		}
+		pkg_spec_parse(&pkgspec, buf);
+		emsg = pkg_spec_is_illegal(&pkgspec);
+		if (emsg)
 			ohshit(_("trigger interest file `%.250s' syntax error; "
 			         "illegal package name `%.250s': %.250s"),
 			       trk_explicit_fn.buf, buf, emsg);
-		pend = findpackage(buf);
-		trig_record_activation(pend, aw, trk_explicit_trig);
+		pend = pkg_spec_find_pkg(&pkgspec, NULL);
+		trig_record_activation(pend, noawait ? NULL : aw,
+		                       trk_explicit_trig);
 	}
+	pkg_spec_reset(&pkgspec);
 }
 
 static void
-trk_explicit_interest_change(const char *trig,  struct pkginfo *pkg, int signum)
+trk_explicit_interest_flush(const char *newfilename, FILE *nf)
+{
+	if (ferror(nf))
+		ohshite(_("unable to write new trigger interest file `%.250s'"),
+		        newfilename);
+	if (fflush(nf))
+		ohshite(_("unable to flush new trigger interest file '%.250s'"),
+		        newfilename);
+	if (fsync(fileno(nf)))
+		ohshite(_("unable to sync new trigger interest file '%.250s'"),
+		        newfilename);
+}
+
+static void
+trk_explicit_interest_commit(const char *newfilename)
+{
+	if (rename(newfilename, trk_explicit_fn.buf))
+		ohshite(_("unable to install new trigger interest file `%.250s'"),
+		        trk_explicit_fn.buf);
+}
+
+static void
+trk_explicit_interest_remove(const char *newfilename)
+{
+	if (unlink(newfilename))
+		ohshite(_("cannot remove `%.250s'"), newfilename);
+	if (unlink(trk_explicit_fn.buf) && errno != ENOENT)
+		ohshite(_("cannot remove `%.250s'"), trk_explicit_fn.buf);
+}
+
+static void
+trk_explicit_interest_change(const char *trig,  struct pkginfo *pkg, int signum,
+                             enum trig_options opts)
 {
 	static struct varbuf newfn;
 	char buf[1024];
 	FILE *nf;
+	bool empty = true;
+	struct pkg_spec pkgspec = PKG_SPEC_INIT(psf_def_native | psf_no_copy |
+	                                        psf_no_check);
 
 	trk_explicit_start(trig);
-	varbufreset(&newfn);
-	varbufprintf(&newfn, "%s/%s.new", triggersdir, trig);
-	varbufaddc(&newfn, 0);
+	varbuf_reset(&newfn);
+	varbuf_printf(&newfn, "%s/%s.new", triggersdir, trig);
 
 	nf = fopen(newfn.buf, "w");
 	if (!nf)
 		ohshite(_("unable to create new trigger interest file `%.250s'"),
 		        newfn.buf);
-	push_cleanup(cu_closefile, ~ehflag_normaltidy, NULL, 0, 1, nf);
+	push_cleanup(cu_closestream, ~ehflag_normaltidy, NULL, 0, 1, nf);
 
 	while (trk_explicit_f && trk_explicit_fgets(buf, sizeof(buf)) >= 0) {
-		if (!strcmp(buf, pkg->name))
+		char *slash = strchr(buf, '/');
+		if (slash)
+			*slash = '\0';
+		pkg_spec_parse(&pkgspec, buf);
+		if (pkg_spec_match_pkg(&pkgspec, pkg, &pkg->installed))
 			continue;
 		fprintf(nf, "%s\n", buf);
+		empty = false;
+	}
+	if (signum > 0) {
+		fprintf(nf, "%s%s\n", pkg_describe(pkg, pdo_foreign),
+		        (opts == trig_noawait) ? "/noawait" : "");
+		empty = false;
 	}
-	if (signum > 0)
-		fprintf(nf, "%s\n", pkg->name);
 
-	if (ferror(nf))
-		ohshite(_("unable to write new trigger interest file `%.250s'"),
-		        newfn.buf);
-	if (fflush(nf))
-		ohshite(_("unable to flush new trigger interest file '%.250s'"),
-		        newfn.buf);
-	if (fsync(fileno(nf)))
-		ohshite(_("unable to sync new trigger interest file '%.250s'"),
-		        newfn.buf);
+	if (!empty)
+		trk_explicit_interest_flush(newfn.buf, nf);
+
 	pop_cleanup(ehflag_normaltidy);
 	if (fclose(nf))
 		ohshite(_("unable to close new trigger interest file `%.250s'"),
 		        newfn.buf);
 
-	if (rename(newfn.buf, trk_explicit_fn.buf))
-		ohshite(_("unable to install new trigger interest file `%.250s'"),
-		        trk_explicit_fn.buf);
+	if (empty)
+		trk_explicit_interest_remove(newfn.buf);
+	else
+		trk_explicit_interest_commit(newfn.buf);
+
+	dir_sync_path(triggersdir);
 }
 
 static const struct trigkindinfo tki_explicit = {
@@ -420,7 +406,7 @@ static const struct trigkindinfo tki_explicit = {
 	.interest_change = trk_explicit_interest_change,
 };
 
-/*---------- file triggers ----------*/
+/*---------- File triggers. ----------*/
 
 static struct {
 	struct trigfileint *head, *tail;
@@ -428,18 +414,20 @@ static struct {
 
 /*
  * Values:
- *  -1: not read,
- *   0: not edited
- *   1: edited
+ *  -1: Not read.
+ *   0: Not edited.
+ *   1: Edited
  */
 static int filetriggers_edited = -1;
 
-/* Called by various people with signum -1 and +1 to mean remove and add
- * and also by trig_file_interests_ensure with signum +2 meaning add
+/*
+ * Called by various people with signum -1 and +1 to mean remove and add
+ * and also by trig_file_interests_ensure() with signum +2 meaning add
  * but die if already present.
  */
 static void
-trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum)
+trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+                         enum trig_options opts)
 {
 	struct filenamenode *fnn;
 	struct trigfileint **search, *tfi;
@@ -456,13 +444,14 @@ trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum)
 		if (tfi->pkg == pkg)
 			goto found;
 
-	/* not found */
+	/* Not found. */
 	if (signum < 0)
 		return;
 
 	tfi = nfmalloc(sizeof(*tfi));
 	tfi->pkg = pkg;
 	tfi->fnn = fnn;
+	tfi->options = opts;
 	tfi->samefile_next = *trigh.namenode_interested(fnn);
 	*trigh.namenode_interested(fnn) = tfi;
 
@@ -470,37 +459,44 @@ trk_file_interest_change(const char *trig, struct pkginfo *pkg, int signum)
 	goto edited;
 
 found:
+	tfi->options = opts;
 	if (signum > 1)
 		ohshit(_("duplicate file trigger interest for filename `%.250s' "
-		         "and package `%.250s'"), trig, pkg->name);
+		         "and package `%.250s'"), trig,
+		       pkg_describe(pkg, pdo_foreign));
 	if (signum > 0)
 		return;
 
-	/* remove it: */
+	/* Remove it: */
 	*search = tfi->samefile_next;
 	LIST_UNLINK_PART(filetriggers, tfi, inoverall.);
 edited:
 	filetriggers_edited = 1;
 }
 
-void
-trig_file_interests_save(void)
+static void
+trig_file_interests_remove(void)
+{
+	if (unlink(triggersfilefile) && errno != ENOENT)
+		ohshite(_("cannot remove `%.250s'"), triggersfilefile);
+}
+
+static void
+trig_file_interests_update(void)
 {
 	struct trigfileint *tfi;
 	FILE *nf;
 
-	if (filetriggers_edited <= 0)
-		return;
-
 	nf = fopen(triggersnewfilefile, "w");
 	if (!nf)
 		ohshite(_("unable to create new file triggers file `%.250s'"),
 		        triggersnewfilefile);
-	push_cleanup(cu_closefile, ~ehflag_normaltidy, NULL, 0, 1, nf);
+	push_cleanup(cu_closestream, ~ehflag_normaltidy, NULL, 0, 1, nf);
 
 	for (tfi = filetriggers.head; tfi; tfi = tfi->inoverall.next)
-		fprintf(nf, "%s %s\n", trigh.namenode_name(tfi->fnn),
-		        tfi->pkg->name);
+		fprintf(nf, "%s %s%s\n", trigh.namenode_name(tfi->fnn),
+		        pkg_describe(tfi->pkg, pdo_foreign),
+		        (tfi->options == trig_noawait) ? "/noawait" : "");
 
 	if (ferror(nf))
 		ohshite(_("unable to write new file triggers file `%.250s'"),
@@ -519,6 +515,20 @@ trig_file_interests_save(void)
 	if (rename(triggersnewfilefile, triggersfilefile))
 		ohshite(_("unable to install new file triggers file as `%.250s'"),
 		        triggersfilefile);
+}
+
+void
+trig_file_interests_save(void)
+{
+	if (filetriggers_edited <= 0)
+		return;
+
+	if (!filetriggers.head)
+		trig_file_interests_remove();
+	else
+		trig_file_interests_update();
+
+	dir_sync_path(triggersdir);
 
 	filetriggers_edited = 0;
 }
@@ -530,6 +540,7 @@ trig_file_interests_ensure(void)
 	char linebuf[1024], *space;
 	struct pkginfo *pkg;
 	const char *emsg;
+	struct pkg_spec pkgspec = PKG_SPEC_INIT(psf_def_native | psf_no_check);
 
 	if (filetriggers_edited >= 0)
 		return;
@@ -542,20 +553,31 @@ trig_file_interests_ensure(void)
 		        triggersfilefile);
 	}
 
-	push_cleanup(cu_closefile, ~0, NULL, 0, 1, f);
+	push_cleanup(cu_closestream, ~0, NULL, 0, 1, f);
 	while (fgets_checked(linebuf, sizeof(linebuf), f, triggersfilefile) >= 0) {
+		char *slash;
+		enum trig_options trig_opts = trig_await;
 		space = strchr(linebuf, ' ');
 		if (!space || linebuf[0] != '/')
 			ohshit(_("syntax error in file triggers file `%.250s'"),
 			       triggersfilefile);
 		*space++ = '\0';
-		if ((emsg = illegal_packagename(space, NULL)))
+
+		slash = strchr(space, '/');
+		if (slash && strcmp("/noawait", slash) == 0) {
+			trig_opts = trig_noawait;
+			*slash = '\0';
+		}
+		pkg_spec_parse(&pkgspec, space);
+		emsg = pkg_spec_is_illegal(&pkgspec);
+		if (emsg)
 			ohshit(_("file triggers record mentions illegal "
 			         "package name `%.250s' (for interest in file "
 				 "`%.250s'): %.250s"), space, linebuf, emsg);
-		pkg = findpackage(space);
-		trk_file_interest_change(linebuf, pkg, +2);
+		pkg = pkg_spec_find_pkg(&pkgspec, NULL);
+		trk_file_interest_change(linebuf, pkg, +2, trig_opts);
 	}
+	pkg_spec_reset(&pkgspec);
 	pop_cleanup(ehflag_normaltidy);
 ok:
 	filetriggers_edited = 0;
@@ -579,7 +601,8 @@ trig_file_activate(struct filenamenode *trig, struct pkginfo *aw)
 
 	for (tfi = *trigh.namenode_interested(trig); tfi;
 	     tfi = tfi->samefile_next)
-		trig_record_activation(tfi->pkg, aw, trigh.namenode_name(trig));
+		trig_record_activation(tfi->pkg, (tfi->options == trig_noawait) ?
+		                       NULL : aw, trigh.namenode_name(trig));
 }
 
 static void
@@ -608,52 +631,54 @@ static const struct trigkindinfo tki_file = {
 	.interest_change = trk_file_interest_change,
 };
 
-/*---------- trigger control info file ----------*/
+/*---------- Trigger control info file. ----------*/
 
 static void
-trig_cicb_interest_change(const char *trig, struct pkginfo *pkg, int signum)
+trig_cicb_interest_change(const char *trig, struct pkginfo *pkg, int signum,
+                          enum trig_options opts)
 {
 	const struct trigkindinfo *tki = trig_classify_byname(trig);
 
 	assert(filetriggers_edited >= 0);
-	tki->interest_change(trig, pkg, signum);
+	tki->interest_change(trig, pkg, signum, opts);
 }
 
 void
-trig_cicb_interest_delete(const char *trig, void *user)
+trig_cicb_interest_delete(const char *trig, void *user, enum trig_options opts)
 {
-	trig_cicb_interest_change(trig, user, -1);
+	trig_cicb_interest_change(trig, user, -1, opts);
 }
 
 void
-trig_cicb_interest_add(const char *trig, void *user)
+trig_cicb_interest_add(const char *trig, void *user, enum trig_options opts)
 {
-	trig_cicb_interest_change(trig, user, +1);
+	trig_cicb_interest_change(trig, user, +1, opts);
 }
 
 void
-trig_cicb_statuschange_activate(const char *trig, void *user)
+trig_cicb_statuschange_activate(const char *trig, void *user,
+                                enum trig_options opts)
 {
 	struct pkginfo *aw = user;
 
 	trig_activate_start(trig);
-	dtki->activate_awaiter(aw);
+	dtki->activate_awaiter((opts == trig_noawait) ? NULL : aw);
 	dtki->activate_done();
 }
 
 static void
 parse_ci_call(const char *file, const char *cmd, trig_parse_cicb *cb,
-              const char *trig, void *user)
+              const char *trig, void *user, enum trig_options opts)
 {
 	const char *emsg;
 
-	emsg = illegal_triggername(trig);
+	emsg = trig_name_is_illegal(trig);
 	if (emsg)
 		ohshit(_("triggers ci file `%.250s' contains illegal trigger "
 		         "syntax in trigger name `%.250s': %.250s"),
 		       file, trig, emsg);
 	if (cb)
-		cb(trig, user);
+		cb(trig, user, opts);
 }
 
 void
@@ -667,11 +692,10 @@ trig_parse_ci(const char *file, trig_parse_cicb *interest,
 	f = fopen(file, "r");
 	if (!f) {
 		if (errno == ENOENT)
-			/* No file is just like an empty one. */
-			return;
+			return; /* No file is just like an empty one. */
 		ohshite(_("unable to open triggers ci file `%.250s'"), file);
 	}
-	push_cleanup(cu_closefile, ~0, NULL, 0, 1, f);
+	push_cleanup(cu_closestream, ~0, NULL, 0, 1, f);
 
 	while ((l = fgets_checked(linebuf, sizeof(linebuf), f, file)) >= 0) {
 		for (cmd = linebuf; cisspace(*cmd); cmd++);
@@ -689,18 +713,22 @@ trig_parse_ci(const char *file, trig_parse_cicb *interest,
 		while (cisspace(*spc))
 			spc++;
 		if (!strcmp(cmd, "interest")) {
-			parse_ci_call(file, cmd, interest, spc, user);
+			parse_ci_call(file, cmd, interest, spc, user, trig_await);
+		} else if (!strcmp(cmd, "interest-noawait")) {
+			parse_ci_call(file, cmd, interest, spc, user, trig_noawait);
 		} else if (!strcmp(cmd, "activate")) {
-			parse_ci_call(file, cmd, activate, spc, user);
+			parse_ci_call(file, cmd, activate, spc, user, trig_await);
+		} else if (!strcmp(cmd, "activate-noawait")) {
+			parse_ci_call(file, cmd, activate, spc, user, trig_noawait);
 		} else {
 			ohshit(_("triggers ci file contains unknown directive `%.250s'"),
 			       cmd);
 		}
 	}
-	pop_cleanup(ehflag_normaltidy); /* fclose */
+	pop_cleanup(ehflag_normaltidy); /* fclose() */
 }
 
-/*---------- Unincorp file incorporation ----------*/
+/*---------- Unincorp file incorporation. ----------*/
 
 static void
 tdm_incorp_trig_begin(const char *trig)
@@ -711,7 +739,11 @@ tdm_incorp_trig_begin(const char *trig)
 static void
 tdm_incorp_package(const char *awname)
 {
-	struct pkginfo *aw = strcmp(awname, "-") ? findpackage(awname) : NULL;
+	struct pkginfo *aw = NULL;
+	if (strcmp(awname, "-")) {
+		struct pkg_spec ps = PKG_SPEC_INIT(psf_def_native);
+		aw = pkg_spec_find_pkg(&ps, awname);
+	}
 
 	dtki->activate_awaiter(aw);
 }
@@ -723,18 +755,27 @@ tdm_incorp_trig_end(void)
 }
 
 static const struct trigdefmeths tdm_incorp = {
-	tdm_incorp_trig_begin,
-	tdm_incorp_package,
-	tdm_incorp_trig_end
+	.trig_begin = tdm_incorp_trig_begin,
+	.package = tdm_incorp_package,
+	.trig_end = tdm_incorp_trig_end
 };
 
 void
-trig_incorporate(enum modstatdb_rw cstatus, const char *admindir)
+trig_incorporate(enum modstatdb_rw cstatus)
 {
-	int ur;
+	enum trigdef_update_status ur;
 	enum trigdef_updateflags tduf;
 
-	trigdef = &tdm_incorp;
+	free(triggersdir);
+	triggersdir = dpkg_db_get_path(TRIGGERSDIR);
+
+	free(triggersfilefile);
+	triggersfilefile = trig_get_filename(triggersdir, "File");
+
+	free(triggersnewfilefile);
+	triggersnewfilefile = trig_get_filename(triggersdir, "File.new");
+
+	trigdef_set_methods(&tdm_incorp);
 	trig_file_interests_ensure();
 
 	tduf = tduf_nolockok;
@@ -744,8 +785,8 @@ trig_incorporate(enum modstatdb_rw cstatus, const char *admindir)
 			tduf |= tduf_writeifenoent;
 	}
 
-	ur = trigdef_update_start(tduf, admindir);
-	if (ur == -1 && cstatus >= msdbrw_write) {
+	ur = trigdef_update_start(tduf);
+	if (ur == tdus_error_no_dir && cstatus >= msdbrw_write) {
 		if (mkdir(triggersdir, 0755)) {
 			if (errno != EEXIST)
 				ohshite(_("unable to create triggers state"
@@ -754,22 +795,22 @@ trig_incorporate(enum modstatdb_rw cstatus, const char *admindir)
 			ohshite(_("unable to set ownership of triggers state"
 			          " directory `%.250s'"), triggersdir);
 		}
-		ur = trigdef_update_start(tduf, admindir);
+		ur = trigdef_update_start(tduf);
 	}
 	switch (ur) {
-	case -2:
+	case tdus_error_empty_deferred:
 		return;
-	case -1:
-	case -3:
+	case tdus_error_no_dir:
+	case tdus_error_no_deferred:
 		if (!trigh.transitional_activate)
 			return;
 	/* Fall through. */
-	case 1:
+	case tdus_no_deferred:
 		trigh.transitional_activate(cstatus);
 		break;
-	case 2:
+	case tdus_ok:
 		/* Read and incorporate triggers. */
-		trigdef_yylex();
+		trigdef_parse();
 		break;
 	default:
 		internerr("unknown trigdef_update_start return value '%d'", ur);
@@ -779,7 +820,7 @@ trig_incorporate(enum modstatdb_rw cstatus, const char *admindir)
 	trigdef_process_done();
 }
 
-/*---------- default hooks ----------*/
+/*---------- Default hooks. ----------*/
 
 struct filenamenode {
 	struct filenamenode *next;
@@ -790,13 +831,14 @@ struct filenamenode {
 static struct filenamenode *trigger_files;
 
 static struct filenamenode *
-th_simple_nn_find(const char *name, int nonew)
+th_simple_nn_find(const char *name, bool nonew)
 {
 	struct filenamenode *search;
 
 	for (search = trigger_files; search; search = search->next)
 		if (!strcmp(search->name, name))
 			return search;
+
 	/* Not found. */
 	if (nonew)
 		return NULL;
@@ -812,7 +854,7 @@ th_simple_nn_find(const char *name, int nonew)
 
 TRIGHOOKS_DEFINE_NAMENODE_ACCESSORS
 
-struct trig_hooks trigh = {
+static struct trig_hooks trigh = {
 	.enqueue_deferred = NULL,
 	.transitional_activate = NULL,
 	.namenode_find = th_simple_nn_find,
@@ -820,3 +862,8 @@ struct trig_hooks trigh = {
 	.namenode_name = th_nn_name,
 };
 
+void
+trig_override_hooks(const struct trig_hooks *hooks)
+{
+	trigh = *hooks;
+}
