@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+=encoding utf8
+
 =head1 NAME
 
 Dpkg::Changelog - base class to implement a changelog parser
@@ -34,6 +36,8 @@ package Dpkg::Changelog;
 use strict;
 use warnings;
 
+our $VERSION = "1.00";
+
 use Dpkg;
 use Dpkg::Gettext;
 use Dpkg::ErrorHandling qw(:DEFAULT report);
@@ -44,8 +48,9 @@ use Dpkg::Index;
 use Dpkg::Version;
 use Dpkg::Vendor qw(run_vendor_hook);
 
+use base qw(Dpkg::Interface::Storable);
+
 use overload
-    '""'  => sub { return $_[0]->output() },
     '@{}' => sub { return $_[0]->{data} };
 
 =over 4
@@ -74,14 +79,6 @@ Parse $filename as a changelog.
 
 =cut
 
-sub load {
-    my ($self, $file) = @_;
-    open(my $fh, "<", $file) or syserr(_g("cannot read %s"), $file);
-    my $ret = $self->parse($fh, $file);
-    close($fh);
-    return $ret;
-}
-
 =item $c->set_options(%opts)
 
 Change the value of some options. "verbose" (defaults to 1) defines
@@ -100,7 +97,7 @@ sub set_options {
 
 =item $c->reset_parse_errors()
 
-Can be used to delete all information about errors ocurred during
+Can be used to delete all information about errors occurred during
 previous L<parse> runs.
 
 =cut
@@ -255,7 +252,7 @@ sub __sanity_check_range {
     }
     if ((defined($r->{since}) and not exists $versions{$r->{since}})) {
         warning(_g("'%s' option specifies non-existing version"), "since");
-        warning(_g("use newest entry that is smaller than the one specified"));
+        warning(_g("use newest entry that is earlier than the one specified"));
         foreach my $v (@versions) {
             if (version_compare_relation($v, REL_LT, $r->{since})) {
                 $r->{since} = $v;
@@ -263,7 +260,7 @@ sub __sanity_check_range {
             }
         }
         if (not exists $versions{$r->{since}}) {
-            # No version was smaller, include all
+            # No version was earlier, include all
             warning(_g("none found, starting from the oldest entry"));
             delete $r->{since};
             $r->{from} = $versions[-1];
@@ -271,7 +268,7 @@ sub __sanity_check_range {
     }
     if ((defined($r->{from}) and not exists $versions{$r->{from}})) {
         warning(_g("'%s' option specifies non-existing version"), "from");
-        warning(_g("use oldest entry that is bigger than the one specified"));
+        warning(_g("use oldest entry that is later than the one specified"));
         my $oldest;
         foreach my $v (@versions) {
             if (version_compare_relation($v, REL_GT, $r->{from})) {
@@ -282,12 +279,12 @@ sub __sanity_check_range {
             $r->{from} = $oldest;
         } else {
             warning(_g("no such entry found, ignoring '%s' parameter"), "from");
-            delete $r->{from}; # No version was bigger
+            delete $r->{from}; # No version was oldest
         }
     }
     if (defined($r->{'until'}) and not exists $versions{$r->{'until'}}) {
         warning(_g("'%s' option specifies non-existing version"), "until");
-        warning(_g("use oldest entry that is bigger than the one specified"));
+        warning(_g("use oldest entry that is later than the one specified"));
         my $oldest;
         foreach my $v (@versions) {
             if (version_compare_relation($v, REL_GT, $r->{'until'})) {
@@ -298,12 +295,12 @@ sub __sanity_check_range {
             $r->{'until'} = $oldest;
         } else {
             warning(_g("no such entry found, ignoring '%s' parameter"), "until");
-            delete $r->{'until'}; # No version was bigger
+            delete $r->{'until'}; # No version was oldest
         }
     }
     if (defined($r->{to}) and not exists $versions{$r->{to}}) {
         warning(_g("'%s' option specifies non-existing version"), "to");
-        warning(_g("use newest entry that is smaller than the one specified"));
+        warning(_g("use newest entry that is earlier than the one specified"));
         foreach my $v (@versions) {
             if (version_compare_relation($v, REL_LT, $r->{to})) {
                 $r->{to} = $v;
@@ -311,7 +308,7 @@ sub __sanity_check_range {
             }
         }
         if (not exists $versions{$r->{to}}) {
-            # No version was smaller
+            # No version was earlier
             warning(_g("no such entry found, ignoring '%s' parameter"), "to");
             delete $r->{to};
         }
@@ -432,6 +429,10 @@ sub abort_early {
     return;
 }
 
+=item $c->save($filename)
+
+Save the changelog in the given file.
+
 =item $c->output()
 
 =item "$c"
@@ -523,7 +524,8 @@ sub dpkg {
     my $f = Dpkg::Control::Changelog->new();
     $f->{Urgency} = $entry->get_urgency() || "unknown";
     $f->{Source} = $entry->get_source() || "unknown";
-    $f->{Version} = $entry->get_version() || "unknown";
+    $f->{Version} = $entry->get_version();
+    $f->{Version} = "unknown" unless defined $f->{Version};
     $f->{Distribution} = join(" ", $entry->get_distributions());
     $f->{Maintainer} = $entry->get_maintainer() || '';
     $f->{Date} = $entry->get_timestamp() || '';
@@ -588,7 +590,8 @@ sub rfc822 {
 	my $f = Dpkg::Control::Changelog->new();
 	$f->{Urgency} = $entry->get_urgency() || "unknown";
 	$f->{Source} = $entry->get_source() || "unknown";
-	$f->{Version} = $entry->get_version() || "unknown";
+	$f->{Version} = $entry->get_version();
+	$f->{Version} = "unknown" unless defined $f->{Version};
 	$f->{Distribution} = join(" ", $entry->get_distributions());
 	$f->{Maintainer} = $entry->get_maintainer() || "";
 	$f->{Date} = $entry->get_timestamp() || "";
@@ -686,7 +689,7 @@ with only one of the options specified.
 =head1 AUTHOR
 
 Frank Lichtenheld, E<lt>frank@lichtenheld.deE<gt>
-Raphael Hertzog, E<lt>hertzog@debian.orgE<gt>
+Raphaël Hertzog, E<lt>hertzog@debian.orgE<gt>
 
 =cut
 1;
