@@ -43,7 +43,7 @@
 #include <dpkg/dpkg-db.h>
 #include <dpkg/arch.h>
 #include <dpkg/pkg-show.h>
-#include <dpkg/myopt.h>
+#include <dpkg/options.h>
 
 #include "filesdb.h"
 #include "infodb.h"
@@ -411,7 +411,7 @@ predeppackage(const char *const *argv)
     pkg->clientdata->istobe= itb_preinstall;
     for (dep= pkg->available.depends; dep; dep= dep->next) {
       if (dep->type != dep_predepends) continue;
-      if (depisok(dep, &vb, NULL, true))
+      if (depisok(dep, &vb, NULL, NULL, true))
         continue;
       /* This will leave dep non-NULL, and so exit the loop. */
       break;
@@ -471,7 +471,7 @@ predeppackage(const char *const *argv)
     pkg->clientdata->istobe= itb_preinstall;
     for (dep= pkg->available.depends; dep; dep= dep->next) {
       if (dep->type != dep_predepends) continue;
-      if (depisok(dep, &vb, NULL, true))
+      if (depisok(dep, &vb, NULL, NULL, true))
         continue;
       /* This will leave dep non-NULL, and so exit the loop. */
       break;
@@ -566,8 +566,8 @@ cmpversions(const char *const *argv)
   };
 
   const struct relationinfo *rip;
-  const char *emsg;
   struct versionrevision a, b;
+  struct dpkg_error err;
   int r;
 
   if (!argv[0] || !argv[1] || !argv[2] || argv[3])
@@ -579,16 +579,24 @@ cmpversions(const char *const *argv)
   if (!rip->string) badusage(_("--compare-versions bad relation"));
 
   if (*argv[0] && strcmp(argv[0],"<unknown>")) {
-    emsg= parseversion(&a,argv[0]);
-    if (emsg)
-      ohshit(_("version '%s' has bad syntax: %s"), argv[0], emsg);
+    if (parseversion(&a, argv[0], &err) < 0) {
+      if (err.type == DPKG_MSG_WARN)
+        warning(_("version '%s' has bad syntax: %s"), argv[0], err.str);
+      else
+        ohshit(_("version '%s' has bad syntax: %s"), argv[0], err.str);
+      dpkg_error_destroy(&err);
+    }
   } else {
     blankversion(&a);
   }
   if (*argv[2] && strcmp(argv[2],"<unknown>")) {
-    emsg= parseversion(&b,argv[2]);
-    if (emsg)
-      ohshit(_("version '%s' has bad syntax: %s"), argv[2], emsg);
+    if (parseversion(&b, argv[2], &err) < 0) {
+      if (err.type == DPKG_MSG_WARN)
+        warning(_("version '%s' has bad syntax: %s"), argv[2], err.str);
+      else
+        ohshit(_("version '%s' has bad syntax: %s"), argv[2], err.str);
+      dpkg_error_destroy(&err);
+    }
   } else {
     blankversion(&b);
   }
