@@ -49,7 +49,7 @@
 #include <dpkg/pkg-show.h>
 #include <dpkg/pkg-spec.h>
 #include <dpkg/path.h>
-#include <dpkg/myopt.h>
+#include <dpkg/options.h>
 
 #include "filesdb.h"
 #include "infodb.h"
@@ -364,8 +364,8 @@ searchfiles(const char *const *argv)
       iterfileend(it);
     }
     if (!found) {
-      fprintf(stderr, _("%s: no path found matching pattern %s.\n"), thisname,
-              thisarg);
+      fprintf(stderr, _("%s: no path found matching pattern %s.\n"),
+              dpkg_get_progname(), thisarg);
       failures++;
       m_output(stderr, _("<standard error>"));
     } else {
@@ -693,8 +693,8 @@ usage(const struct cmdinfo *ci, const char *value)
   exit(0);
 }
 
-const char thisname[]= "dpkg-query";
-const char printforhelp[]= N_("Use --help for help about querying packages.");
+static const char printforhelp[] = N_(
+"Use --help for help about querying packages.");
 
 static const char *admindir;
 
@@ -725,8 +725,9 @@ int main(int argc, const char *const *argv) {
   bindtextdomain(PACKAGE, LOCALEDIR);
   textdomain(PACKAGE);
 
+  dpkg_set_progname("dpkg-query");
   standard_startup();
-  myopt(&argv, cmdinfos);
+  myopt(&argv, cmdinfos, printforhelp);
 
   admindir = dpkg_db_set_dir(admindir);
 
