@@ -12,7 +12,7 @@
  * Domain.
  *
  * Changes by Ben Collins <bcollins@debian.org>, added --chuid, --background
- * and --make-pidfile options, placed in public domain aswell.
+ * and --make-pidfile options, placed in public domain as well.
  *
  * Port to OpenBSD by Sontri Tomo Huynh <huynh.29@osu.edu>
  *                 and Andreas Schuldei <andreas@schuldei.org>
@@ -116,8 +116,18 @@
 #endif
 
 #if defined(OSLinux)
-/* This comes from TASK_COMM_LEN defined in Linux's include/linux/sched.h. */
+/* This comes from TASK_COMM_LEN defined in Linux' include/linux/sched.h. */
 #define PROCESS_NAME_SIZE 15
+#elif defined(OSsunos)
+#define PROCESS_NAME_SIZE 15
+#elif defined(OSDarwin)
+#define PROCESS_NAME_SIZE 16
+#elif defined(OSNetBSD)
+#define PROCESS_NAME_SIZE 16
+#elif defined(OSOpenBSD)
+#define PROCESS_NAME_SIZE 16
+#elif defined(OSFreeBSD)
+#define PROCESS_NAME_SIZE 19
 #endif
 
 #if defined(SYS_ioprio_set) && defined(linux)
@@ -137,11 +147,17 @@ enum {
 	IOPRIO_CLASS_IDLE,
 };
 
+enum action_code {
+	action_none,
+	action_start,
+	action_stop,
+	action_status,
+};
+
+static enum action_code action;
 static int testmode = 0;
 static int quietmode = 0;
 static int exitnodo = 1;
-static int start = 0;
-static int stop = 0;
 static int background = 0;
 static int mpidfile = 0;
 static int signal_nr = SIGTERM;
@@ -172,6 +188,14 @@ static struct stat exec_stat;
 static struct proc_stat_list *procset = NULL;
 #endif
 
+/* LSB Init Script process status exit codes. */
+enum status_code {
+	status_ok = 0,
+	status_dead_pidfile = 1,
+	status_dead_lockfile = 2,
+	status_dead = 3,
+	status_unknown = 4,
+};
 
 struct pid_list {
 	struct pid_list *next;
@@ -233,7 +257,10 @@ fatal(const char *format, ...)
 	else
 		fprintf(stderr, "\n");
 
-	exit(2);
+	if (action == action_status)
+		exit(status_unknown);
+	else
+		exit(2);
 }
 
 static void *
@@ -370,6 +397,7 @@ usage(void)
 "Commands:\n"
 "  -S|--start -- <argument> ...  start a program and pass <arguments> to it\n"
 "  -K|--stop                     stop a program\n"
+"  -T|--status                   get the program status\n"
 "  -H|--help                     print help information\n"
 "  -V|--version                  print version\n"
 "\n"
@@ -388,7 +416,7 @@ usage(void)
 "  -a|--startas <pathname>       program to start (default is <executable>)\n"
 "  -r|--chroot <directory>       chroot to <directory> before starting\n"
 "  -d|--chdir <directory>        change to <directory> (default is /)\n"
-"  -N|--nicelevel <incr>         add incr to the process's nice level\n"
+"  -N|--nicelevel <incr>         add incr to the process' nice level\n"
 "  -P|--procsched <policy[:prio]>\n"
 "                                use <policy> with <prio> for the kernel\n"
 "                                  process scheduler (default prio is 0)\n"
@@ -415,8 +443,16 @@ usage(void)
 "The IO scheduler <class> can be one of:\n"
 "  real-time, best-effort or idle\n"
 "\n"
-"Exit status:  0 = done      1 = nothing done (=> 0 if --oknodo)\n"
-"              3 = trouble   2 = with --retry, processes wouldn't die\n");
+"Exit status:\n"
+"  0 = done\n"
+"  1 = nothing done (=> 0 if --oknodo)\n"
+"  2 = with --retry, processes would not die\n"
+"  3 = trouble\n"
+"Exit status with --status:\n"
+"  0 = program is running\n"
+"  1 = program is not running and the pid file exists\n"
+"  3 = program is not running\n"
+"  4 = unable to determine status\n");
 }
 
 static void
@@ -433,7 +469,11 @@ badusage(const char *msg)
 	if (msg)
 		fprintf(stderr, "%s: %s\n", progname, msg);
 	fprintf(stderr, "Try '%s --help' for more information.\n", progname);
-	exit(3);
+
+	if (action == action_status)
+		exit(status_unknown);
+	else
+		exit(3);
 }
 
 struct sigpair {
@@ -708,6 +748,18 @@ parse_schedule(const char *schedule_str)
 	}
 }
 
+static void
+set_action(enum action_code new_action)
+{
+	if (action == new_action)
+		return;
+
+	if (action != action_none)
+		badusage("only one command can be specified");
+
+	action = new_action;
+}
+
 static void
 parse_options(int argc, char * const *argv)
 {
@@ -715,6 +767,7 @@ parse_options(int argc, char * const *argv)
 		{ "help",	  0, NULL, 'H'},
 		{ "stop",	  0, NULL, 'K'},
 		{ "start",	  0, NULL, 'S'},
+		{ "status",	  0, NULL, 'T'},
 		{ "version",	  0, NULL, 'V'},
 		{ "startas",	  1, NULL, 'a'},
 		{ "name",	  1, NULL, 'n'},
@@ -748,7 +801,7 @@ parse_options(int argc, char * const *argv)
 
 	for (;;) {
 		c = getopt_long(argc, argv,
-		                "HKSVa:n:op:qr:s:tu:vx:c:N:P:I:k:bmR:g:d:",
+		                "HKSVTa:n:op:qr:s:tu:vx:c:N:P:I:k:bmR:g:d:",
 		                longopts, NULL);
 		if (c == -1)
 			break;
@@ -757,10 +810,13 @@ parse_options(int argc, char * const *argv)
 			usage();
 			exit(0);
 		case 'K':  /* --stop */
-			stop = 1;
+			set_action(action_stop);
 			break;
 		case 'S':  /* --start */
-			start = 1;
+			set_action(action_start);
+			break;
+		case 'T':  /* --status */
+			set_action(action_status);
 			break;
 		case 'V':  /* --version */
 			do_version();
@@ -859,8 +915,8 @@ parse_options(int argc, char * const *argv)
 			badusage("umask value must be a positive number");
 	}
 
-	if (start == stop)
-		badusage("need one of --start or --stop");
+	if (action == action_none)
+		badusage("need one of --start or --stop or --status");
 
 	if (!execname && !pidfile && !userspec && !cmdname)
 		badusage("need at least one of --exec, --pidfile, --user or --name");
@@ -875,15 +931,14 @@ parse_options(int argc, char * const *argv)
 	if (!startas)
 		startas = execname;
 
-	if (start && !startas)
+	if (action == action_start && !startas)
 		badusage("--start needs --exec or --startas");
 
 	if (mpidfile && pidfile == NULL)
 		badusage("--make-pidfile requires --pidfile");
 
-	if (background && !start)
+	if (background && action != action_start)
 		badusage("--background is only relevant with --start");
-
 }
 
 #if defined(OSHurd)
@@ -1003,7 +1058,7 @@ pid_is_user(pid_t pid, uid_t uid)
 	struct proc_stat *ps;
 
 	ps = get_proc_stat(pid, PSTAT_OWNER_UID);
-	return ps && proc_stat_owner_uid(ps) == uid;
+	return ps && (uid_t)proc_stat_owner_uid(ps) == uid;
 }
 #elif defined(OShpux)
 static bool
@@ -1111,7 +1166,7 @@ pid_is_cmd(pid_t pid, const char *name)
 	end_argv_0_p = strchr(*pid_argv_p, ' ');
 	if (end_argv_0_p == NULL)
 		/* There seems to be no space, so we have the command
-		 * allready in its desired form. */
+		 * already in its desired form. */
 		start_argv_0_p = *pid_argv_p;
 	else {
 		/* Tests indicate that this never happens, since
@@ -1147,57 +1202,70 @@ pid_is_running(pid_t pid)
 }
 #endif
 
-static void
+static enum status_code
 pid_check(pid_t pid)
 {
 #if defined(OSLinux) || defined(OShpux)
 	if (execname && !pid_is_exec(pid, &exec_stat))
-		return;
+		return status_dead;
 #elif defined(HAVE_KVM_H)
 	if (execname && !pid_is_exec(pid, execname))
-		return;
+		return status_dead;
 #elif defined(OSHurd) || defined(OSFreeBSD) || defined(OSNetBSD)
 	/* Let's try this to see if it works. */
 	if (execname && !pid_is_cmd(pid, execname))
-		return;
+		return status_dead;
 #endif
 	if (userspec && !pid_is_user(pid, user_id))
-		return;
+		return status_dead;
 	if (cmdname && !pid_is_cmd(pid, cmdname))
-		return;
-	if (start && !pid_is_running(pid))
-		return;
+		return status_dead;
+	if (action == action_start && !pid_is_running(pid))
+		return status_dead;
+
 	pid_list_push(&found, pid);
+
+	return status_ok;
 }
 
-static void
+static enum status_code
 do_pidfile(const char *name)
 {
 	FILE *f;
 	static pid_t pid = 0;
 
-	if (pid) {
-		pid_check(pid);
-		return;
-	}
+	if (pid)
+		return pid_check(pid);
 
 	f = fopen(name, "r");
 	if (f) {
+		enum status_code pid_status;
+
 		if (fscanf(f, "%d", &pid) == 1)
-			pid_check(pid);
+			pid_status = pid_check(pid);
+		else
+			pid_status = status_unknown;
 		fclose(f);
-	} else if (errno != ENOENT)
+
+		if (pid_status == status_dead)
+			return status_dead_pidfile;
+		else
+			return pid_status;
+	} else if (errno == ENOENT)
+		return status_dead;
+	else
 		fatal("unable to open pidfile %s", name);
 }
 
 #if defined(OSLinux) || defined (OSsunos)
-static void
+static enum status_code
 do_procinit(void)
 {
 	DIR *procdir;
 	struct dirent *entry;
 	int foundany;
 	pid_t pid;
+	enum status_code prog_status = status_dead;
 
 	procdir = opendir("/proc");
 	if (!procdir)
@@ -1205,14 +1273,21 @@ do_procinit(void)
 
 	foundany = 0;
 	while ((entry = readdir(procdir)) != NULL) {
+		enum status_code pid_status;
+
 		if (sscanf(entry->d_name, "%d", &pid) != 1)
 			continue;
 		foundany++;
-		pid_check(pid);
+
+		pid_status = pid_check(pid);
+		if (pid_status < prog_status)
+			prog_status = pid_status;
 	}
 	closedir(procdir);
 	if (!foundany)
 		fatal("nothing in /proc - not mounted?");
+
+	return prog_status;
 }
 #elif defined(OSHurd)
 static int
@@ -1222,49 +1297,63 @@ check_proc_stat(struct proc_stat *ps)
 	return 0;
 }
 
-static void
+static enum status_code
 do_procinit(void)
 {
 	if (!procset)
 		init_procset();
 
 	proc_stat_list_for_each(procset, check_proc_stat);
+
+	if (found)
+		return status_ok;
+	else
+		return status_dead;
 }
 #elif defined(OShpux)
-static void
+static enum status_code
 do_procinit(void)
 {
 	struct pst_status pst[10];
 	int i, count;
 	int idx = 0;
+	enum status_code prog_status = status_dead;
 
 	while ((count = pstat_getproc(pst, sizeof(pst[0]), 10, idx)) > 0) {
-		for (i = 0; i < count; i++)
-			pid_check(pst[i].pst_pid);
+		enum status_code pid_status;
+
+		for (i = 0; i < count; i++) {
+			pid_status = pid_check(pst[i].pst_pid);
+			if (pid_status < prog_status)
+				prog_status = pid_status;
+		}
 		idx = pst[count - 1].pst_idx + 1;
 	}
+
+	return prog_status;
 }
 #elif defined(HAVE_KVM_H)
-static void
+static enum status_code
 do_procinit(void)
 {
 	/* Nothing to do. */
+	return status_unknown;
 }
 #endif
 
-static void
+static enum status_code
 do_findprocs(void)
 {
 	pid_list_free(&found);
 
 	if (pidfile)
-		do_pidfile(pidfile);
+		return do_pidfile(pidfile);
 	else
-		do_procinit();
+		return do_procinit();
 }
 
 static void
-do_stop(int sig_num, int quiet, int *n_killed, int *n_notkilled, int retry_nr)
+do_stop(int sig_num, int *n_killed, int *n_notkilled)
 {
 	struct pid_list *p;
 
@@ -1294,15 +1383,23 @@ do_stop(int sig_num, int quiet, int *n_killed, int *n_notkilled, int retry_nr)
 			(*n_notkilled)++;
 		}
 	}
-	if (quietmode < 0 && killed) {
-		printf("Stopped %s (pid", what_stop);
-		for (p = killed; p; p = p->next)
-			printf(" %d", p->pid);
-		putchar(')');
-		if (retry_nr > 0)
-			printf(", retry #%d", retry_nr);
-		printf(".\n");
-	}
+}
+
+static void
+do_stop_summary(int retry_nr)
+{
+	struct pid_list *p;
+
+	if (quietmode >= 0 || !killed)
+		return;
+
+	printf("Stopped %s (pid", what_stop);
+	for (p = killed; p; p = p->next)
+		printf(" %d", p->pid);
+	putchar(')');
+	if (retry_nr > 0)
+		printf(", retry #%d", retry_nr);
+	printf(".\n");
 }
 
 static void
@@ -1346,7 +1443,7 @@ do_stop_timeout(int timeout, int *n_killed, int *n_notkilled)
 		if (timercmp(&before, &stopat, >))
 			return false;
 
-		do_stop(0, 1, n_killed, n_notkilled, 0);
+		do_stop(0, n_killed, n_notkilled);
 		if (!*n_killed)
 			return true;
 
@@ -1419,7 +1516,8 @@ run_stop_schedule(void)
 	retry_nr = 0;
 
 	if (schedule == NULL) {
-		do_stop(signal_nr, quietmode, &n_killed, &n_notkilled, 0);
+		do_stop(signal_nr, &n_killed, &n_notkilled);
+		do_stop_summary(0);
 		if (n_notkilled > 0 && quietmode <= 0)
 			printf("%d pids were not killed\n", n_notkilled);
 		if (n_killed)
@@ -1437,7 +1535,8 @@ run_stop_schedule(void)
 			position = value;
 			goto reposition;
 		case sched_signal:
-			do_stop(value, quietmode, &n_killed, &n_notkilled, retry_nr++);
+			do_stop(value, &n_killed, &n_notkilled);
+			do_stop_summary(retry_nr++);
 			if (!n_killed)
 				return finish_stop_schedule(anykilled);
 			else
@@ -1463,6 +1562,7 @@ run_stop_schedule(void)
 int
 main(int argc, char **argv)
 {
+	enum status_code prog_status;
 	int devnull_fd = -1;
 	gid_t rgid;
 	uid_t ruid;
@@ -1506,6 +1606,7 @@ main(int argc, char **argv)
 		struct group *gr = getgrnam(changegroup);
 		if (!gr)
 			fatal("group '%s' not found", changegroup);
+		changegroup = gr->gr_name;
 		runas_gid = gr->gr_gid;
 	}
 	if (changeuser) {
@@ -1518,6 +1619,7 @@ main(int argc, char **argv)
 			pw = getpwnam(changeuser);
 		if (!pw)
 			fatal("user '%s' not found", changeuser);
+		changeuser = pw->pw_name;
 		runas_uid = pw->pw_uid;
 		if (changegroup == NULL) {
 			/* Pass the default group of this user. */
@@ -1528,12 +1630,15 @@ main(int argc, char **argv)
 			setenv("HOME", pw->pw_dir, 1);
 	}
 
-	if (stop) {
+	if (action == action_stop) {
 		int i = run_stop_schedule();
 		exit(i);
 	}
 
-	do_findprocs();
+	prog_status = do_findprocs();
+
+	if (action == action_status)
+		exit(prog_status);
 
 	if (found) {
 		if (quietmode <= 0)
